use crate::dynamic_data::{
    Battle, BattleParty, BattleRandom, BattleResult, BattleSide, ChoiceQueue, DynamicLibrary, Pokemon, PokemonParty,
};
use crate::script_implementations::wasm::export_registry::register;
use crate::script_implementations::wasm::extern_ref::{ExternRef, VecExternRef};
use crate::script_implementations::wasm::script_resolver::WebAssemblyEnv;
use crate::StringKey;
use wasmer::FunctionEnvMut;

register! {
    fn battle_get_pokemon_per_side(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> u8 {
        battle.value_func(&env).unwrap().pokemon_per_side()
    }

    fn battle_get_parties(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> VecExternRef<BattleParty> {
        VecExternRef::new(env.data().data().as_ref(), battle.value_func(&env).unwrap().parties())
    }

    fn battle_get_choice_queue(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> ExternRef<ChoiceQueue> {
        let queue = battle.value_func(&env).unwrap().current_turn_queue().read();
        if let Some(queue) = queue.as_ref() {
            ExternRef::func_new(&env, queue)
        } else {
            ExternRef::null()
        }
    }

    fn battle_get_library(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> ExternRef<dyn DynamicLibrary> {
        ExternRef::func_new(&env, &battle.value_func_arc(&env).unwrap().library().clone())
    }

    fn battle_get_sides(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> VecExternRef<BattleSide> {
        VecExternRef::new(env.data().data().as_ref(), battle.value_func(&env).unwrap().sides())
    }

    fn battle_get_random(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> ExternRef<BattleRandom> {
        ExternRef::func_new(&env, battle.value_func(&env).unwrap().random())
    }

    fn battle_get_weather_name(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> ExternRef<StringKey> {
        let weather = battle.value_func(&env).unwrap().weather_name().unwrap();
        if let Some(weather) = weather {
            ExternRef::func_new(&env, &weather)
        } else {
            ExternRef::null()
        }
    }

    fn battle_find_party_for_pokemon(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
        pokemon: ExternRef<Pokemon>
    ) -> ExternRef<BattleParty> {
        let battle = battle.value_func(&env).unwrap();
        let pokemon = pokemon.value_func(&env).unwrap();
        for party in battle.parties() {
            if party.party().has_pokemon(pokemon) {
                return ExternRef::func_new(&env, party);
            }
        }
        ExternRef::null()
    }

    fn battle_get_pokemon(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
        side: u8, index: u8
    ) -> ExternRef<Pokemon> {
        let battle = battle.value_func(&env).unwrap();
        let pokemon = battle.get_pokemon(side, index);
        if let Some(pokemon) = pokemon {
            ExternRef::func_new(&env, &pokemon)
        } else {
            ExternRef::null()
        }
    }

    fn battle_get_can_flee(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> u8 {
        if battle.value_func(&env).unwrap().can_flee() { 1 } else { 0 }
    }

    fn battle_get_number_of_sides(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> u8 {
        battle.value_func(&env).unwrap().number_of_sides()
    }

    fn battle_get_has_ended(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> u8 {
        if battle.value_func(&env).unwrap().has_ended() { 1 } else { 0 }
    }

    fn battle_get_has_ended_conclusively(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> u8 {
        if battle.value_func(&env).unwrap().result().is_conclusive() { 1 } else { 0 }
    }

    fn battle_get_winning_side(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> u8 {
        if let BattleResult::Conclusive(result) = battle.value_func(&env).unwrap().result() {
            result
        } else {
            0
        }
    }

    fn battle_get_current_turn(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle: ExternRef<Battle>,
    ) -> u32 {
        battle.value_func(&env).unwrap().current_turn()
    }

    fn battle_party_get_party(
        env: FunctionEnvMut<WebAssemblyEnv>,
        battle_party: ExternRef<BattleParty>,
    ) -> ExternRef<PokemonParty> {
        ExternRef::func_new(&env, battle_party.value_func(&env).unwrap().party().as_ref())
    }
}
